import aiohttp
import asyncio
import ssl
from typing import Optional


class Connection:
    def __init__(
        self,
        host: str,
        port: int,
        path: str,
        ssl_context: Optional[ssl.SSLContext] = None,
    ):
        """Create a connection to the server.

        Args:
            host (str): The hostname of the server.
            port (int): The port of the server.
            path (str): The path to the server.
            ssl_context (ssl.SSLContext, optional): The SSL context. Defaults to None.
        """
        self.host = host
        self.port = port
        self.path = path
        self.protocol = "wss" if port == 443 else "ws"
        self.url = "{}://{}:{}{}".format(self.protocol, self.host, self.port, self.path)
        self.ws = None
        self.ssl_context = ssl_context

    async def connect(self):
        """Connect to the server."""
        self.session = aiohttp.ClientSession()
        print("connecting...")
        if self.ssl_context is not None:
            self.ws = await self.session.ws_connect(self.url, ssl=self.ssl_context)
        else:
            self.ws = await self.session.ws_connect(self.url)

    async def send(self, message: str):
        """Send a message to the server.

        Args:
            message (str): The message to send.
        """
        await self.ws.send_str(message)

    async def receive(self):
        """Receive a message from the server.

        Returns:
            str: The message received.
        """
        return await self.ws.receive()

    async def close(self):
        """Close the connection to the server."""
        await self.ws.close()

    def __str__(self):
        """Return a string representation of the connection.

        Returns:
            str: The string representation of the connection.
        """
        return "Connection: {}".format(self.url)

    def __repr__(self):
        """Return a representation of the connection.

        Returns:
            str: The representation of the connection.
        """
        return self.__str__()



import asyncio
import configparser
import importlib
import os
import sys
import ssl
from typing import Optional

import aiohttp

from pyshowdown import connection, message


class Client:
    def __init__(
        self,
        host: str,
        port: int,
        path: str,
        ssl_context: Optional[ssl.SSLContext] = None,
    ):
        """Client class constructor.

        Args:
            host (str): Hostname or IP address of the server.
            port (int): The port number of the server.
            path (str): The path to the server.
            ssl_context (ssl.SSLContext, optional): The SSL context. Defaults to None.
        """
        self.conn = connection.Connection(host, port, path, ssl_context=ssl_context)
        self.connected = False
        self.load_config()
        self.plugins = []
        self.load_plugins()

    def load_config(self):
        """Load config from config.ini."""
        self.config = configparser.ConfigParser()
        self.config.read("config.ini")
        self.username = self.config["user"]["username"]
        self.password = self.config["user"]["password"]
        self.plugin_dir = self.config["user"].get("plugin_dir", "system")
        self.plugin_list = self.config["user"].get("plugins").split(",")

    async def connect(self):
        """Connect to the server."""
        await self.conn.connect()

    async def keep_connected(self):
        """Keeps the client connected to the server."""
        self.connected = False
        timeout = 1
        while not self.connected:
            try:
                await asyncio.sleep(timeout)
                await self.connect()
                self.connected = True
                timeout = 1
                await self.receive_forever()
            except Exception as e:
                print(e)
                timeout += 1

    async def close(self):
        """Close the connection."""
        await self.conn.close()

    async def send(self, room: str, message: str):
        """Sends message to the server.

        Args:
            message (str): The message to send.
        """
        await self.conn.send(f"{room}|{message}")
    
    async def send_pm(self, user: str, message: str):
        """Sends a private message to the user.

        Args:
            user (str): The user to send the message to.
            message (str): The message to send.
        """
        await self.send("", f"/w {user}, {message}")

    async def receive(self) -> str:
        """Receives data from the server.

        Returns:
            str: The data received.
        """
        return await self.conn.receive()

    async def receive_forever(self):
        """Receives data from the server forever."""
        async for ws_message in self.conn.ws:
            if ws_message.type == aiohttp.WSMsgType.TEXT:
                print(ws_message)
                await self.handle_message(ws_message.data)
        self.connected = False

    def load_plugins(self):
        """Loads all the plugins from the directory set in config.

        It should first import them, then instantiate the class
        which is a subclass of BasePlugin.
        """
        print("Loading plugins...")

        # always load the system plugins
        sys.path.append(os.path.join(os.path.dirname(__file__), "plugins"))

        if self.plugin_dir != "system":
            sys.path.append(self.plugin_dir)

        for plugin in self.plugin_list:
            try:
                plugin_module = importlib.import_module(plugin)
                plugins = plugin_module.setup(self)
                for plugin in plugins:
                    self.plugins.append(plugin)
            except Exception as e:
                print("Error loading plugin {}: {}".format(plugin, e))

    async def handle_message(self, msg_str: str):
        """Handles a message from the server.

        Iterates through all the loaded plugins, determines whether
        any of them can handle the message, and if so, calls the
        response method of the plugin.

        Args:
            msg_str (str): The message received.
        """
        print(msg_str)
        m = message.Message("", msg_str)

        for plugin in self.plugins:
            matched = await plugin.match(m)
            if matched:
                resp = await plugin.response(m)
                if resp:
                    if m.type == "pm":
                        await self.send_pm(m.sender, resp)
                    else:
                        await self.send(m.room, resp)
    
    async def join(self, room: str):
        """Joins the given room.

        Args:
            room (str): The room to join.
        """
        await self.send("", f"/join {room}")
    
    async def leave(self, room: str):
        """Leaves the given room.

        Args:
            room (str): The room to leave.
        """
        await self.send(room, "/leave")

    def __str__(self) -> str:
        """Returns a string representation of the client.

        Returns:
            str: The string representation of the client.
        """
        return "Client({}, {}, {})".format(
            self.conn.host, self.conn.port, self.conn.path
        )

    def __repr__(self) -> str:
        """Returns a representation of the client.

        Returns:
            str: The representation of the client.
        """
        return self.__str__()



from pyshowdown.client import Client


class BasePlugin:
    def __init__(self, client: Client):
        """Initializes the plugin.

        Args:
            client (Client): A reference to the client.
        """
        self.client = client

    async def match(self, message):
        """Returns True if the message is a match for the plugin.

        Args:
            message (Message): The message to check.

        Raises:
            NotImplementedError: Always, since this is a base class.
        """
        raise NotImplementedError()

    async def response(self, message):
        """Returns the response for the message.

        Args:
            message (Message): The message to respond to.

        Raises:
            NotImplementedError: Always, since this is a base class.
        """
        raise NotImplementedError()



import json


class Message:
    def __init__(self, room: str, message_str: str):
        """Initialize a Message object.

        Args:
            room (str): The room the message was sent to.
            message_str (str): The raw message.
        """
        self.room = room
        self.message_str = message_str
        self.parse_message()

    def parse_message(self):
        """Parse the message and store the attributes."""
        info = self.message_str.split("|")
        self.type = info[1]

        # Room initialization
        if self.type == "init":
            self.roomtype = info[2]

        elif self.type == "title":
            self.title = info[2]

        elif self.type == "users":
            users = info[2].split(",")
            self.usercount = int(users.pop(0))

            self.users = []

            # this should probably be refactored to use User objects
            for u in users:
                rank = u[0]
                u = u[1:]
                if "@" in u:
                    name, status = u.split("@")
                else:
                    name, status = u, ""

                if status and status[0] == "!":
                    away = True
                    status = status[1:]
                else:
                    away = False

                self.users.append((rank, name, status, away))

        # Room messages
        elif self.type == "html":
            self.html = info[2]

        elif self.type == "uhtml":
            self.name = info[2]
            self.html = info[3]

        elif self.type == "uhtmlchange":
            self.name = info[2]
            self.html = info[3]

        elif self.type in ["j", "J", "join"]:
            self.type = "join"
            self.user = info[2]

        elif self.type in ["l", "L", "leave"]:
            self.type = "leave"
            self.user = info[2]

        elif self.type in ["n", "N", "name"]:
            self.type = "name"
            self.user = info[2]
            self.oldid = info[3]

        elif self.type in ["c", "chat"]:
            self.type = "chat"
            self.timestamp = None
            self.user = info[2]
            self.message = info[3]

        elif self.type == ":":
            self.type = "timestamp"
            self.timestamp = int(info[2])

        elif self.type == "c:":
            self.type = "chat"
            self.timestamp = int(info[2])
            self.user = info[3]
            self.message = info[4]

        elif self.type == "battle":
            self.roomid = info[2]
            self.user1 = info[3]
            self.user2 = info[4]

        # Global messages
        elif self.type == "popup":
            self.message = info[2]

        elif self.type == "pm":
            self.sender = info[2]
            self.receiver = info[3]
            self.message = info[4]

        elif self.type == "usercount":
            self.usercount = int(info[2])

        elif self.type == "nametaken":
            self.username = info[2]
            self.message = info[3]

        elif self.type == "challstr":
            self.challstr = "|".join(info[2:])

        elif self.type == "updateuser":
            self.user = info[2]
            self.named = True if info[3] == "1" else False
            self.avatar = info[4]
            self.settings = json.loads(info[5])

        elif self.type == "formats":
            self.formats = parse_formats("|".join(info[2:]))

        elif self.type == "updatesearch":
            self.json = json.loads("|".join(info[2:]))

        elif self.type == "updatechallenges":
            self.json = json.loads("|".join(info[2:]))

        elif self.type == "queryresponse":
            self.query_type = info[2]

            json_str = "|".join(info[3:])
            self.json_data = json.loads(json_str)

            if self.query_type == "savereplay":
                # override the relevant room, since queryresponse messages
                # are sent as global messages
                self.room = self.json_data["id"]

        elif self.type == "raw":
            self.data = info[2]

        # Battle messages
        elif self.type == "win":
            self.winner = info[2]

        elif self.type == "player":
            self.player = info[2] if len(info) > 2 else None
            self.username = info[3] if len(info) > 3 else None
            self.avatar = info[4] if len(info) > 4 else None
            self.rating = int(info[5]) if len(info) > 5 and info[5] else None

        else:
            # TODO: Handle other message types
            pass

    def __str__(self) -> str:
        """Return a string representation of the message.

        Returns:
            str: A string representation of the message.
        """
        return f"<Message: {self.message_str}>"

    def __repr__(self) -> str:
        """Return a representation of the message.

        Returns:
            str: A representation of the message.
        """
        return self.__str__()


def parse_formats(format_str: str) -> dict:
    """Parse a format message and return a list of formats.

    Args:
        format_str (str): The format message.

    Returns:
        dict: A dictionary of sections, formats, and some additional
            info about the formats.
    """
    results = {}
    split_str = format_str.split("|")
    in_section = False
    section_name = ""

    for item in split_str:
        if not item:
            continue
        if in_section:
            section_name = item
            results[section_name] = {}
            in_section = False
        elif item[0] == ",":
            # this is a section
            in_section = True
            section_number = int(item[1:])
        else:
            # this is a format
            name, rule_num = item.split(",")
            rule_num = int(rule_num, 16)
            rules = []

            if rule_num & 1:
                rules.append("Requires Team")
            if rule_num & 2:
                rules.append("Available for Search")
            if rule_num & 4:
                rules.append("Available for Challenge")
            if rule_num & 8:
                rules.append("Available for Tournaments")
            if rule_num & 16:
                rules.append("Level 50")
            results[section_name][name] = rules

    return results



import json

import aiohttp

from pyshowdown.plugins.plugin import BasePlugin
from pyshowdown.message import Message


class ChallstrHandler(BasePlugin):
    async def match(self, message: Message) -> bool:
        """Returns true if the message is a challstr.

        Args:
            message (Message): The message to check.

        Returns:
            bool: True if the message is a challstr, False otherwise.
        """
        return message.type == "challstr"

    async def response(self, message: Message) -> None:
        """Responds to a challstr message and logs in.

        Args:
            message (Message): The message to respond to.
        """
        print("Got challstr! Trying to log in...")
        base_url = "https://play.pokemonshowdown.com/action.php"
        data = {
            "act": "login",
            "name": self.client.username,
            "pass": self.client.password,
            "challstr": message.challstr,
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(base_url, data=data) as resp:
                result = await resp.text()
                self.client.cookies = resp.cookies
                print("----------")
                print(self.client.cookies)
                print("----------")

                # strip the leading [
                result = result[1:]
                result = json.loads(result)

                await self.client.send(
                    "", "/trn {},0,{}".format(self.client.username, result["assertion"])
                )


def setup(client) -> list:
    """Creates an instance of the ChallstrHandler plugin and returns it.

    Args:
        client (Client): A reference to the client.

    Returns:
        list: A list containing the ChallstrHandler plugin.
    """
    return [ChallstrHandler(client)]
